from asyncio import iscoroutine
from peewee import ForeignKeyField
from playhouse.shortcuts import _clone_set


async def model_to_dict(model, recurse=True, backrefs=False, only=None,
                        exclude=None, seen=None, extra_attrs=None,
                        fields_from_query=None, max_depth=None):
    """
    Convert a model instance (and any related objects) to a dictionary.
    :param bool recurse: Whether foreign-keys should be recursed.
    :param bool backrefs: Whether lists of related objects should be recursed.
    :param only: A list (or set) of field instances indicating which fields
        should be included.
    :param exclude: A list (or set) of field instances that should be
        excluded from the dictionary.
    :param list extra_attrs: Names of model instance attributes or methods
        that should be included.
    :param SelectQuery fields_from_query: Query that was source of model. Take
        fields explicitly selected by the query and serialize them.
    :param int max_depth: Maximum depth to recurse, value <= 0 means no max.
    """
    max_depth = -1 if max_depth is None else max_depth
    if max_depth == 0:
        recurse = False

    only = _clone_set(only)
    extra_attrs = _clone_set(extra_attrs)

    if fields_from_query is not None:
        for item in fields_from_query._select:
            if isinstance(item, Field):
                only.add(item)
            elif isinstance(item, Node) and item._alias:
                extra_attrs.add(item._alias)

    data = {}
    exclude = _clone_set(exclude)
    seen = _clone_set(seen)
    exclude |= seen
    model_class = type(model)

    for field in model._meta.declared_fields:
        if field in exclude or (only and (field not in only)):
            continue

        field_data = model._data.get(field.name)
        if isinstance(field, ForeignKeyField) and recurse:
            if field_data:
                seen.add(field)
                rel_obj = getattr(model, field.name)
                if iscoroutine(rel_obj):
                    rel_obj = await rel_obj
                field_data = await model_to_dict(
                    rel_obj,
                    recurse=recurse,
                    backrefs=backrefs,
                    only=only,
                    exclude=exclude,
                    seen=seen,
                    max_depth=max_depth - 1)
            else:
                field_data = None

        data[field.name] = field_data

    if extra_attrs:
        for attr_name in extra_attrs:
            attr = getattr(model, attr_name)
            if callable(attr):
                data[attr_name] = attr()
            else:
                data[attr_name] = attr

    if backrefs and recurse:
        for related_name, foreign_key in model._meta.reverse_rel.items():
            descriptor = getattr(model_class, related_name)
            if descriptor in exclude or foreign_key in exclude:
                continue
            if only and (descriptor not in only) and (foreign_key not in only):
                continue

            accum = []
            exclude.add(foreign_key)
            related_query = getattr(
                model,
                related_name + '_prefetch',
                getattr(model, related_name))

            async for rel_obj in related_query:
                accum.append(await model_to_dict(
                    rel_obj,
                    recurse=recurse,
                    backrefs=backrefs,
                    only=only,
                    exclude=exclude,
                    max_depth=max_depth - 1))

            data[related_name] = accum

    return data



__all__ = ['parser']
import platform
import getpass
from collections import namedtuple
from urllib.parse import urlparse,ParseResult,parse_qsl

from .error import *

from typing import Dict,Optional,Tuple,Union,Any




Default = namedtuple('Default',['user','password','host','port'])

if platform.system() == 'Darwin':
    user = getpass.getuser()
    password = ''
elif platform.system() == 'windows':
    user = 'postgres'
    password = 'postgres'
else:
    user = 'postgres'
    password = ''

DEFAULT_INFO = {
    'mysql':Default(user='root',password='',host='localhost',port='3306'),
    'postgresql':Default(user=user,password=password,host='localhost',port='5432')
}

def check_scheme(scheme:Optional[str])->str:

    SCHEMES = ('mysql','postgresql')
    if scheme:
        if scheme.lower() in SCHEMES:
            return scheme.lower()
        else:
            raise InvalidURI("unknow database")
    else:
        raise InvalidURI("must have a database uri")

def check_netloc(parse_result:ParseResult,scheme:str)->Tuple[Optional[str],Optional[str],Optional[str],Optional[str]]:
    default = DEFAULT_INFO.get(scheme)

    user = parse_result.username or default.user
    password = parse_result.password or default.password
    host = parse_result.hostname or default.host
    if parse_result.port:
        port = str(parse_result.port)
    else:
        port = default.port
    return user,password,host,port

def check_path(path:str)->str:
    if path:
        return path.replace("/","")
    else:
        raise InvalidURI("need to point out the db's name")

def check_query(query:str)->Any:
    if query:
        return dict(parse_qsl(query))
    else:
        return None


def parser(uri:str)->Dict[str,str]:
    parse_result = urlparse(uri)
    scheme = check_scheme(parse_result.scheme)
    usr,password,host,port = check_netloc(parse_result,scheme)
    db = check_path(parse_result.path)
    query = check_query(parse_result.query)

    result = dict(
    scheme=scheme,
    username = usr,
    password = password,
    host = host,
    port = int(port),
    database = db
    )
    if query:
        result.update(**query)
    return result



from peewee import QueryResultWrapper, ExtQueryResultWrapper
from peewee import TuplesQueryResultWrapper, DictQueryResultWrapper
from peewee import ModelQueryResultWrapper, AggregateQueryResultWrapper
from peewee import NaiveQueryResultWrapper
from aitertools import aiter
class AioResultIterator(object):

    def __init__(self, qrw):
        self.qrw = qrw
        self._idx = 0

    async def __anext__(self):
        if self._idx < self.qrw._ct:
            obj = self.qrw._result_cache[self._idx]
        elif not self.qrw._populated:
            obj = await self.qrw.iterate()
            self.qrw._result_cache.append(obj)
            self.qrw._ct += 1
        else:
            raise StopAsyncIteration
        self._idx += 1
        return obj


class AioQueryResultWrapper(QueryResultWrapper):

    async def __aiter__(self):
    #def __aiter__(self):
        if self._populated:
            return iter(self._result_cache)
        else:
            return AioResultIterator(self)

    def __await__(self):
        return self.all().__await__()

    async def all(self):
        if not self._initialized:
            self.initialize(self.cursor.description)
            self._initialized = True

        rows = [self.process_row(row) for row in await self.cursor.fetchall()]
        return rows
        self._result_cache = rows
        self._populated = True
        self._idx = self._ct = len(rows)
        return rows

    @property
    def count(self):
        raise NotImplementedError()
        # self.fill_cache()
        # return self._ct

    def __len__(self):
        raise NotImplementedError()
        # return self.count

    async def iterate(self):
        row = await self.cursor.fetchone()
        if not row:
            self._populated = True
            if not getattr(self.cursor, 'name', None):
                # WHY?
                try:
                    await self.cursor.close()
                except TypeError as ts:
                    self.cursor.close()
                except:
                    raise
            raise StopAsyncIteration
        elif not self._initialized:
            self.initialize(self.cursor.description)
            self._initialized = True
        return self.process_row(row)

    def iterator(self):
        raise NotImplementedError()
        # while True:
        #     yield self.iterate()

    async def __anext__(self):
        if self._idx < self._ct:
            inst = self._result_cache[self._idx]
            self._idx += 1
            return inst
        elif self._populated:
            raise StopAsyncIteration

        obj = await self.iterate()
        self._result_cache.append(obj)
        self._ct += 1
        self._idx += 1
        return obj

    async def fill_cache(self, n=None):
        n = n or float('Inf')
        if n < 0:
            raise ValueError('Negative values are not supported.')
        self._idx = self._ct
        while not self._populated and (n > self._ct):
            try:
                await self.__anext__()
            except StopIteration:
                break


class AioExtQueryResultWrapper(AioQueryResultWrapper,
                               ExtQueryResultWrapper):
    pass
    #async def iterator(self):
    #    print("123")
    #    return aiter(self)
        #raise NotImplementedError()
        # while True:
        #     yield self.iterate()


class AioTuplesQueryResultWrapper(AioExtQueryResultWrapper,
                                  TuplesQueryResultWrapper):
    pass


class AioNaiveQueryResultWrapper(AioExtQueryResultWrapper,
                                 NaiveQueryResultWrapper):
    pass


class AioDictQueryResultWrapper(AioExtQueryResultWrapper,
                                DictQueryResultWrapper):
    pass


class AioModelQueryResultWrapper(AioQueryResultWrapper,
                                 ModelQueryResultWrapper):
    pass


class AioAggregateQueryResultWrapper(AioModelQueryResultWrapper,
                                     AggregateQueryResultWrapper):

    async def iterate(self):
        if self._row:
            row = self._row.pop()
        else:
            row = await self.cursor.fetchone()

        if not row:
            self._populated = True
            if not getattr(self.cursor, 'name', None):
                await self.cursor.close()
            raise StopAsyncIteration
        elif not self._initialized:
            self.initialize(self.cursor.description)
            self._initialized = True

        def _get_pk(instance):
            if instance._meta.composite_key:
                return tuple([
                    instance._data[field_name]
                    for field_name in instance._meta.primary_key.field_names])
            return instance._get_pk_value()

        identity_map = {}
        _constructed = self.construct_instances(row)
        primary_instance = _constructed[self.model]
        for model_or_alias, instance in _constructed.items():
            identity_map[model_or_alias] = OrderedDict()
            identity_map[model_or_alias][_get_pk(instance)] = instance

        model_data = self.read_model_data(row)
        while True:
            cur_row = await self.cursor.fetchone()
            if cur_row is None:
                break

            duplicate_models = set()
            cur_row_data = self.read_model_data(cur_row)
            for model_class, data in cur_row_data.items():
                if model_data[model_class] == data:
                    duplicate_models.add(model_class)

            if not duplicate_models:
                self._row.append(cur_row)
                break

            different_models = self.all_models - duplicate_models

            new_instances = self.construct_instances(cur_row, different_models)
            for model_or_alias, instance in new_instances.items():
                # Do not include any instances which are comprised solely of
                # NULL values.
                all_none = True
                for value in instance._data.values():
                    if value is not None:
                        all_none = False
                if not all_none:
                    identity_map[model_or_alias][_get_pk(instance)] = instance

        stack = [self.model]
        instances = [primary_instance]
        while stack:
            current = stack.pop()
            if current not in self.join_meta:
                continue

            for join in self.join_meta[current]:
                try:
                    metadata, attr = self.source_to_dest[current][join.dest]
                except KeyError:
                    continue

                if metadata.is_backref or metadata.is_self_join:
                    for instance in identity_map[current].values():
                        setattr(instance, attr, [])

                    if join.dest not in identity_map:
                        continue

                    for pk, inst in identity_map[join.dest].items():
                        if pk is None:
                            continue
                        try:
                            # XXX: if no FK exists, unable to join.
                            joined_inst = identity_map[current][
                                inst._data[metadata.foreign_key.name]]
                        except KeyError:
                            continue

                        getattr(joined_inst, attr).append(inst)
                        instances.append(inst)
                elif attr:
                    if join.dest not in identity_map:
                        continue

                    for pk, instance in identity_map[current].items():
                        # XXX: if no FK exists, unable to join.
                        joined_inst = identity_map[join.dest][
                            instance._data[metadata.foreign_key.name]]
                        setattr(
                            instance,
                            metadata.foreign_key.name,
                            joined_inst)
                        instances.append(joined_inst)

                stack.append(join.dest)

        for instance in instances:
            instance._prepare_instance()

        return primary_instance



from unittest import TestCase
import asyncio
from aioorm import AioModel, AioMySQLDatabase,AioPostgreSQLDatabase
from aioorm.utils import aiodump_csv,aioload_csv
from peewee import (ForeignKeyField, IntegerField, CharField,
                    DateTimeField, TextField, PrimaryKeyField,Proxy,
                    SQL)

from io import StringIO
import aiofiles
from itertools import count
from pathlib import Path
import os
from aitertools import alist
db = Proxy()

c = count()

class User_csv_f(AioModel):
    id = PrimaryKeyField()
    name = CharField(max_length=25)
    age = IntegerField()
    sex = CharField(max_length=1)

    def __repr__(self):
        return '{self.id},{self.name},{self.age},{self.sex}'.format(self=self)

    class Meta:
        database = db

class User_csv(AioModel):
    name = CharField(max_length=25)
    age = IntegerField()
    sex = CharField(max_length=1)

    class Meta:
        database = db

async def test_csv_dump(loop):
    await db.connect(loop)

    assert await User_csv.table_exists() is False
    await User_csv.create_table()
    assert await User_csv.table_exists() is True
    iq = User_csv.insert_many([
        {'name': 'u1',"age":18,'sex':'f'},
        {'name': 'u2',"age":17,'sex':'f'},
        {'name': 'u3',"age":16,'sex':'m'},
        {'name': 'u4',"age":15,'sex':'f'}])

    await iq.execute()
    p = Path(__file__).absolute().parent
    filepath = p.joinpath('source/{}_user_out.csv'.format(str(next(c))))
    query = User_csv.select().order_by(User_csv.id)
    await aiodump_csv(query,str(filepath))
    with StringIO(newline='') as f:
        f = await aiodump_csv(query,f)
        async with aiofiles.open(str(filepath)) as ff:
            r1 = await ff.read()
            r2 = f.getvalue()
            assert r1 == r2

    await User_csv.drop_table()
    assert await User_csv.table_exists() is False

    await db.close()

async def test_csv_load(loop):
    await db.connect(loop)

    p = Path(__file__).absolute().parent
    filepath = p.joinpath('source/user.csv')
    filepath1 = p.joinpath('source/0_user_out.csv')
    head = "id,name,age,sex"+os.linesep
    content = """1,u1,18,f
2,u2,17,f
3,u3,16,m
4,u4,15,f
"""
    with StringIO() as f:
        f.write(head+content)
        await aioload_csv(User_csv_f, f,pk_in_csv=True)
        all_cols = SQL('*')
        query0 = User_csv_f.select(all_cols).order_by(User_csv_f.id)
        user0 = await alist(query0)
        s = (os.linesep).join([str(i) for i in user0]).strip()
        r = content.strip()
        assert s == r
    await User_csv_f.drop_table()

    async with aiofiles.open(str(filepath)) as ff:
        er = await aioload_csv(User_csv_f, ff,pk_in_csv=False)
        all_cols = SQL('*')
        query0 = User_csv_f.select(all_cols).order_by(User_csv_f.id)
        user0 = await alist(query0)
        s = (os.linesep).join([str(i) for i in user0]).strip()
        r = content.strip()
        assert s == r

    await User_csv_f.drop_table()

    await aioload_csv(User_csv_f, '0_user_out.csv',pk_in_csv=True)
    all_cols = SQL('*')
    query0 = User_csv_f.select(all_cols).order_by(User_csv_f.id)
    user0 = await alist(query0)
    s = (os.linesep).join([str(i) for i in user0]).strip()
    r = content.strip()
    assert s == r


    await User_csv_f.drop_table()

    assert await User_csv_f.table_exists() is False

    await db.close()

class ModelMysqlTest(TestCase):
        @classmethod
        def setUpClass(cls, *args, **kwargs):
            """Configure database managers, create test tables.
            """
            database = AioMySQLDatabase('test', host='localhost', port=3306,
                       user='root', password='hsz881224')
            db.initialize(database)
            cls.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(cls.loop)
        @classmethod
        def tearDownClass(cls, *args, **kwargs):
            """Remove all test tables and close connections.
            """
            cls.loop.close()

        def test_csv_dump(self):
            self.loop.run_until_complete(test_csv_dump(self.loop))

        def test_csv_load(self):
            self.loop.run_until_complete(test_csv_load(self.loop))



class ModelPostgreSQLTest(TestCase):
        @classmethod
        def setUpClass(cls, *args, **kwargs):
            """Configure database managers, create test tables.
            """
            database = AioPostgreSQLDatabase('test_ext', host='localhost', port=5432,
                      user='huangsizhe', password='')
            db.initialize(database)
            cls.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(cls.loop)
        @classmethod
        def tearDownClass(cls, *args, **kwargs):
            """Remove all test tables and close connections.
            """
            cls.loop.close()

        def test_csv_dump(self):
            self.loop.run_until_complete(test_csv_dump(self.loop))

        def test_csv_load(self):
            self.loop.run_until_complete(test_csv_load(self.loop))



from peewee import ExecutionContext, Using, _atomic, transaction, savepoint
from typing import Callable

class _aio_callable_context_manager:
    __slots__ = ()
    def __call__(self, fn:Callable)->Callable:
        @wraps(fn)
        async def inner(*args, **kwargs):
            async with self:
                return fn(*args, **kwargs)
        return inner


class AioExecutionContext(_aio_callable_context_manager, ExecutionContext):

    async def __aenter__(self):
        async with self.database._conn_lock:
            self.database.push_execution_context(self)
            self.connection = await self.database._connect(
                self.database.database,
                **self.database.connect_kwargs)
            if self.with_transaction:
                self.txn = self.database.transaction()
                await self.txn.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        async with self.database._conn_lock:
            if self.connection is None:
                self.database.pop_execution_context()
            else:
                try:
                    if self.with_transaction:
                        if not exc_type:
                            self.txn.commit(False)
                        await self.txn.__aexit__(exc_type, exc_val, exc_tb)
                finally:
                    self.database.pop_execution_context()
                    await self.database._close(self.connection)


class AioUsing(AioExecutionContext, Using):

    async def __aenter__(self):
        self._orig = []
        for model in self.models:
            self._orig.append(model._meta.database)
            model._meta.database = self.database
        return await super(Using, self).__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await super(Using, self).__aexit__(exc_type, exc_val, exc_tb)
        for i, model in enumerate(self.models):
            model._meta.database = self._orig[i]


class _aio_atomic(_aio_callable_context_manager, _atomic):

    async def __aenter__(self):
        if self.db.transaction_depth() == 0:
            self.context_manager = self.db.transaction(self.transaction_type)
        else:
            self.context_manager = self.db.savepoint()
        return await self.context_manager.__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.context_manager.__aexit__(exc_type, exc_val, exc_tb)


class aio_transaction(_aio_callable_context_manager, transaction):

    async def _begin(self):
        if self.transaction_type:
            await self.db.begin(self.transaction_type)
        else:
            await self.db.begin()

    async def commit(self, begin:bool=True):
        await self.db.commit()
        if begin:
            await self._begin()

    async def rollback(self, begin=True):
        await self.db.rollback()
        if begin:
            await self._begin()

    async def __aenter__(self):
        self.autocommit = self.db.get_autocommit()
        self.db.set_autocommit(False)
        if self.db.transaction_depth() == 0: await self._begin()
        self.db.push_transaction(self)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        try:
            if exc_type:
                await self.rollback(False)
            elif self.db.transaction_depth() == 1:
                try:
                    await self.commit(False)
                except:
                    await self.rollback(False)
                    raise
        finally:
            self.db.set_autocommit(self.autocommit)
            self.db.pop_transaction()


class aio_savepoint(_aio_callable_context_manager, savepoint):

    async def _execute(self, query):
        await self.db.execute_sql(query, require_commit=False)

    async def _begin(self):
        await self._execute('SAVEPOINT %s;' % self.quoted_sid)

    async def commit(self, begin=True):
        await self._execute('RELEASE SAVEPOINT %s;' % self.quoted_sid)
        if begin: await self._begin()

    async def rollback(self):
        await self._execute('ROLLBACK TO SAVEPOINT %s;' % self.quoted_sid)

    async def __aenter__(self):
        self.autocommit = self.db.get_autocommit()
        self.db.set_autocommit(False)
        await self._begin()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        try:
            if exc_type:
                await self.rollback()
            else:
                try:
                    await self.commit(begin=False)
                except:
                    await self.rollback()
                    raise
        finally:
            self.db.set_autocommit(self.autocommit)



import asyncio

from peewee import Database, ExceptionWrapper, basestring
from peewee import sort_models_topologically, merge_dict
from peewee import OperationalError
from peewee import (RESULTS_NAIVE, RESULTS_TUPLES, RESULTS_DICTS,
                    RESULTS_AGGREGATE_MODELS, RESULTS_MODELS)

from .context import _aio_atomic, aio_transaction, aio_savepoint
from .result import (AioNaiveQueryResultWrapper, AioModelQueryResultWrapper,
                     AioTuplesQueryResultWrapper, AioDictQueryResultWrapper,
                     AioAggregateQueryResultWrapper)


class AioDatabase(Database):

    def __init__(self, database, autocommit=True,
                 fields=None, ops=None, autorollback=False,
                 **connect_kwargs):
        self._closed = True
        self._conn_pool = None

        self.connect_kwargs = {}
        self.init(database, **connect_kwargs)

        self.autocommit = autocommit
        self.autorollback = autorollback
        self.use_speedups = False

        self.field_overrides = merge_dict(self.field_overrides, fields or {})
        self.op_overrides = merge_dict(self.op_overrides, ops or {})
        self.exception_wrapper = ExceptionWrapper(self.exceptions)

        # 用于保持连接
        self._auto_task = None

    def is_closed(self):
         return self._closed

    async def get_conn(self):
        if self._closed:
            with self.exception_wrapper:
                await self.connect()

        return await self._conn_pool.acquire()

    async def get_cursor(self):
        raise NotImplementedError()

    async def _close(self, conn):
        await self._conn_pool.release(conn)

    async def close(self):
        if self.deferred:
            raise Exception('Error, database not properly initialized '
                            'before closing connection')
        with self.exception_wrapper:
            if not self._closed and self._conn_pool:
                self._conn_pool.close()
                self._closed = True
                await self._conn_pool.wait_closed()

    async def connect(self, loop=None):
        if self.deferred:
            raise OperationalError('Database has not been initialized')
        if not self._closed:
            raise OperationalError('Connection already open')
        self._conn_pool = await self._create_connection(loop=loop)
        self._closed = False

        with self.exception_wrapper:
            self.initialize_connection(self._conn_pool)

    async def _create_connection(self, loop=None):
        with self.exception_wrapper:
            return await self._connect(self.database, loop=loop,
                                       **self.connect_kwargs)

    async def execute_sql(self, sql, params=None, require_commit=True):
        with self.exception_wrapper:
            conn = await self.get_conn()
            try:
                cursor = await conn.cursor()
                await cursor.execute(sql, params or ())
                # TODO: MIGHT CLOSE THE CURSOR FROM RESULT WRAPPER...
                await cursor.close()
            except Exception:
                if self.get_autocommit() and self.autorollback:
                    await conn.rollback()
                raise
            else:
                if require_commit and self.get_autocommit():
                    await conn.commit()
            finally:
                await self._close(conn)

        return cursor

    def get_result_wrapper(self, wrapper_type):
        if wrapper_type == RESULTS_NAIVE:
            return AioNaiveQueryResultWrapper
        elif wrapper_type == RESULTS_MODELS:
            return AioModelQueryResultWrapper
        elif wrapper_type == RESULTS_TUPLES:
            return AioTuplesQueryResultWrapper
        elif wrapper_type == RESULTS_DICTS:
            return AioDictQueryResultWrapper
        elif wrapper_type == RESULTS_AGGREGATE_MODELS:
            return AioAggregateQueryResultWrapper
        else:
            return AioNaiveQueryResultWrapper

    async def cancel(self, timeout=None):
        raise NotImplementedError

    def set_autocommit(self, autocommit):
        raise NotImplementedError

    def get_autocommit(self):
        raise NotImplementedError

    def transaction(self, transaction_type=None):
        raise NotImplementedError

    def savepoint(self, sid=None):
        raise NotImplementedError

    def atomic(self, transaction_type=None):
        raise NotImplementedError

    async def commit(self):
        raise NotImplementedError

    async def rollback(self):
        raise NotImplementedError

    async def get_tables(self, schema=None):
        raise NotImplementedError

    async def get_indexes(self, table, schema=None):
        raise NotImplementedError

    async def get_columns(self, table, schema=None):
        raise NotImplementedError

    async def get_primary_keys(self, table, schema=None):
        raise NotImplementedError

    async def get_foreign_keys(self, table, schema=None):
        raise NotImplementedError

    async def sequence_exists(self, seq):
        raise NotImplementedError

    async def create_table(self, model_class, safe=False):
        qc = self.compiler()
        return await self.execute_sql(*qc.create_table(model_class, safe))

    async def create_tables(self, models, safe=False):
        await create_model_tables(models, fail_silently=safe)

    async def create_index(self, model_class, fields, unique=False):
        qc = self.compiler()
        if not isinstance(fields, (list, tuple)):
            raise ValueError('Fields passed to "create_index" must be a list '
                             'or tuple: "%s"' % fields)
        fobjs = [
            model_class._meta.fields[f] if isinstance(f, basestring) else f
            for f in fields]
        return await self.execute_sql(*qc.create_index(model_class, fobjs, unique))

    async def drop_index(self, model_class, fields, safe=False):
        qc = self.compiler()
        if not isinstance(fields, (list, tuple)):
            raise ValueError('Fields passed to "drop_index" must be a list '
                             'or tuple: "%s"' % fields)
        fobjs = [
            model_class._meta.fields[f] if isinstance(f, basestring) else f
            for f in fields]
        return await self.execute_sql(*qc.drop_index(model_class, fobjs, safe))

    async def create_foreign_key(self, model_class, field, constraint=None):
        qc = self.compiler()
        return await self.execute_sql(*qc.create_foreign_key(
            model_class, field, constraint))

    async def create_sequence(self, seq):
        if self.sequences:
            qc = self.compiler()
            return await self.execute_sql(*qc.create_sequence(seq))

    async def drop_table(self, model_class, fail_silently=False, cascade=False):
        qc = self.compiler()
        if cascade and not self.drop_cascade:
            raise ValueError('Database does not support DROP TABLE..CASCADE.')
        return await self.execute_sql(*qc.drop_table(
            model_class, fail_silently, cascade))

    async def drop_tables(self, models, safe=False, cascade=False):
        await drop_model_tables(models, fail_silently=safe, cascade=cascade)

    async def truncate_table(self, model_class, restart_identity=False,
                             cascade=False):
        qc = self.compiler()
        return await self.execute_sql(*qc.truncate_table(
            model_class, restart_identity, cascade))

    async def truncate_tables(self, models, restart_identity=False,
                              cascade=False):
        for model in reversed(sort_models_topologically(models)):
            await model.truncate_table(restart_identity, cascade)

    async def drop_sequence(self, seq):
        if self.sequences:
            qc = self.compiler()
            return await self.execute_sql(*qc.drop_sequence(seq))

    def extract_date(self, date_part, date_field):
        return fn.EXTRACT(Clause(date_part, R('FROM'), date_field))

    def truncate_date(self, date_part, date_field):
        return fn.DATE_TRUNC(date_part, date_field)

    def default_insert_clause(self, model_class):
        return SQL('DEFAULT VALUES')

    def get_noop_sql(self):
        return 'SELECT 0 WHERE 0'

    def get_binary_type(self):
        return binary_construct


async def create_model_tables(models, **create_table_kwargs):
    """Create tables for all given models (in the right order)."""
    for m in sort_models_topologically(models):
        await m.create_table(**create_table_kwargs)

async def drop_model_tables(models, **drop_table_kwargs):
    """Drop tables for all given models (in the right order)."""
    for m in reversed(sort_models_topologically(models)):
        await m.drop_table(**drop_table_kwargs)



from peewee import CharField, TextField, ForeignKeyField, CompositeKey,Proxy
from aioorm import AioModel, AioMySQLDatabase, model_to_dict, AioPostgreSQLDatabase
import asyncio
from unittest import TestCase
db = Proxy()



class User(AioModel):
    username = CharField()

    class Meta:
        database = db


class Note(AioModel):
    user = ForeignKeyField(User, related_name='notes')
    text = TextField()

    class Meta:
        database = db


class Tag(AioModel):
    tag = CharField()

    class Meta:
        database = db


class NoteTag(AioModel):
    note = ForeignKeyField(Note)
    tag = ForeignKeyField(Tag)

    class Meta:
        database = db
        primary_key = CompositeKey('note', 'tag')


async def simple_recurse_t(loop):
    await db.connect(loop)
    await db.create_tables([User, Note], safe=True)

    user = await User.create(username='test')
    note = await Note.create(user=user, text='note-1')

    result = await model_to_dict(note)
    expected = {
        'id': note.id,
        'text': note.text,
        'user': {
            'id': user.id,
            'username': user.username
        }
    }
    assert result == expected

    result = await model_to_dict(note, recurse=False)
    expected = {
        'id': note.id,
        'text': note.text,
        'user': user.id
    }
    assert result == expected

    await db.drop_tables([User, Note], safe=True)
    await db.close()


async def model_to_dict_t(loop):
    await db.connect(loop)
    await db.create_tables([User, Note, Tag, NoteTag], safe=True)

    user = await User.create(username='peewee')
    n0, n1, n2 = [await Note.create(user=user, text='n%s' % i)
                  for i in range(3)]
    t0, tx = [await Tag.create(tag=t) for t in ('t0', 'tx')]
    await NoteTag.create(note=n0, tag=t0)
    await NoteTag.create(note=n0, tag=tx)
    await NoteTag.create(note=n1, tag=tx)

    data = await model_to_dict(user, recurse=True, backrefs=True)
    expected = {
        'id': user.id,
        'username': 'peewee',
        'notes': [
            {'id': n0.id, 'text': 'n0', 'notetag_set': [
                {'tag': {'tag': 't0', 'id': t0.id}},
                {'tag': {'tag': 'tx', 'id': tx.id}},
            ]},
            {'id': n1.id, 'text': 'n1', 'notetag_set': [
                {'tag': {'tag': 'tx', 'id': tx.id}},
            ]},
            {'id': n2.id, 'text': 'n2', 'notetag_set': []},
        ]
    }
    assert data == expected

    data = await model_to_dict(user, recurse=True, backrefs=True, max_depth=2)

    await db.drop_tables([User, Note, Tag, NoteTag], safe=True)
    await db.close()


class ShortCutMysqlTest(TestCase):
        @classmethod
        def setUpClass(cls, *args, **kwargs):
            """Configure database managers, create test tables.
            """
            database = AioMySQLDatabase('test', host='localhost', port=3306,
                       user='root', password='hsz881224')
            db.initialize(database)
            cls.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(cls.loop)
        @classmethod
        def tearDownClass(cls, *args, **kwargs):
            """Remove all test tables and close connections.
            """
            cls.loop.close()

        def test_simple_recurse(self):
            self.loop.run_until_complete(simple_recurse_t(self.loop))

        def test_model_to_dict(self):
            self.loop.run_until_complete(model_to_dict_t(self.loop))

class ShortCutPostgreSQLTest(TestCase):
        @classmethod
        def setUpClass(cls, *args, **kwargs):
            """Configure database managers, create test tables.
            """
            database = AioPostgreSQLDatabase('test_ext', host='localhost', port=5432,
                      user='huangsizhe', password='')
            db.initialize(database)
            cls.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(cls.loop)
        @classmethod
        def tearDownClass(cls, *args, **kwargs):
            """Remove all test tables and close connections.
            """
            cls.loop.close()

        def test_simple_recurse(self):
            self.loop.run_until_complete(simple_recurse_t(self.loop))

        def test_model_to_dict(self):
            self.loop.run_until_complete(model_to_dict_t(self.loop))



from peewee import Query, RawQuery, SelectQuery, NoopSelectQuery
from peewee import CompoundSelect, DeleteQuery, UpdateQuery, InsertQuery
from peewee import _WriteQuery

RESULTS_NAIVE = 1
RESULTS_MODELS = 2
RESULTS_TUPLES = 3
RESULTS_DICTS = 4
RESULTS_AGGREGATE_MODELS = 5
RESULTS_NAMEDTUPLES = 6
class AioQuery(Query):

    async def execute(self):
        raise NotImplementedError

    async def _execute(self):
        sql, params = self.sql()
        return await self.database.execute_sql(sql, params,
                                               self.require_commit)

    async def scalar(self, as_tuple=False, convert=False):
        if convert:
            row = await self.tuples().first()
        else:
            cursor = await self._execute()
            row = await cursor.fetchone()
        if row and not as_tuple:
            return row[0]
        else:
            return row

    def __await__(self):
        return self.execute().__await__()

    def __iter__(self):
        raise NotImplementedError()
        #return await self

    #async def __aiter__(self):
    async def __aiter__(self):
        return await self.execute()


class AioRawQuery(AioQuery, RawQuery):

    async def execute(self):
        if self._qr is None:
            if self._tuples:
                QRW = self.database.get_result_wrapper(RESULTS_TUPLES)
            elif self._dicts:
                QRW = self.database.get_result_wrapper(RESULTS_DICTS)
            else:
                QRW = self.database.get_result_wrapper(RESULTS_NAIVE)
            self._qr = QRW(self.model_class, await self._execute(), None)
        return self._qr


class AioSelectQuery(AioQuery, SelectQuery):

    def compound_op(operator):
        def inner(self, other):
            supported_ops = self.model_class._meta.database.compound_operations
            if operator not in supported_ops:
                raise ValueError(
                    'Your database does not support %s' % operator)
            return AioCompoundSelect(self.model_class, self, operator, other)
        return inner

    async def aggregate(self, aggregation=None, convert=True):
        return await self._aggregate(aggregation).scalar(convert=convert)

    async def count(self, clear_limit=False):
        if self._distinct or self._group_by or self._limit or self._offset:
            return self.wrapped_count(clear_limit=clear_limit)

        # defaults to a count() of the primary key
        return await self.aggregate(convert=False) or 0

    async def wrapped_count(self, clear_limit=False):
        clone = self.order_by()
        if clear_limit:
            clone._limit = clone._offset = None

        sql, params = clone.sql()
        wrapped = 'SELECT COUNT(1) FROM (%s) AS wrapped_select' % sql
        rq = self.model_class.raw(wrapped, *params)
        return await rq.scalar() or 0

    async def all(self):
        qr = await self.execute()
        return await qr.all()

    def __await__(self):
        return self.all().__await__()

    async def exists(self):
        clone = self.paginate(1, 1)
        clone._select = [SQL('1')]
        return bool(await clone.scalar())

    async def get(self):
        clone = self.paginate(1, 1)
        try:
            qr = await clone.execute()
            return await qr.__anext__()
        except StopIteration:
            raise self.model_class.DoesNotExist(
                'Instance matching query does not exist:\nSQL: %s\nPARAMS: %s'
                % self.sql())

    async def peek(self, n=1):
        res = await self.execute()
        await res.fill_cache(n)
        models = res._result_cache[:n]
        if models:
            return models[0] if n == 1 else models

    async def first(self, n=1):
        if self._limit != n:
            self._limit = n
            self._dirty = True
        return await self.peek(n=n)

    def sql(self):
        return self.compiler().generate_select(self)

    async def execute(self):
        if self._dirty or self._qr is None:
            model_class = self.model_class
            query_meta = self.get_query_meta()
            ResultWrapper = self._get_result_wrapper()
            cursor = await self._execute()
            self._qr = ResultWrapper(model_class, cursor, query_meta)
            self._dirty = False
            return self._qr
        else:
            return self._qr

    async def __anext__(self):
        res = await self.execute()
        a = await res.__anext__()
        return a
    def __aiter__(self):
        return self

    async def iterator(self):
        return self.__aiter__()
        #raise NotImplementedError()


    def __getitem__(self, value):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def __hash__(self):
        return id(self)


class AioNoopSelectQuery(AioSelectQuery, NoopSelectQuery):
    pass


class AioCompoundSelect(AioSelectQuery, CompoundSelect):
    _node_type = 'compound_select_query'

    async def count(self, clear_limit=False):
        return await self.wrapped_count(clear_limit=clear_limit)


class _AioWriteQuery(AioQuery, _WriteQuery):

    async def _execute_with_result_wrapper(self):
        ResultWrapper = self.get_result_wrapper()
        meta = (self._returning, {self.model_class: []})
        self._qr = ResultWrapper(self.model_class, await self._execute(), meta)
        return self._qr


class AioUpdateQuery(_AioWriteQuery, UpdateQuery):

    async def execute(self):
        if self._returning is not None and self._qr is None:
            return await self._execute_with_result_wrapper()
        elif self._qr is not None:
            return self._qr
        else:
            return self.database.rows_affected(await self._execute())

    async def __aiter__(self):
        if not self.model_class._meta.database.returning_clause:
            raise ValueError('UPDATE queries cannot be iterated over unless '
                             'they specify a RETURNING clause, which is not '
                             'supported by your database.')
        return await self.execute()


class AioInsertQuery(_AioWriteQuery, InsertQuery):

    async def _insert_with_loop(self):
        id_list = []
        last_id = None
        return_id_list = self._return_id_list
        for row in self._rows:
            last_id = await (InsertQuery(self.model_class, row)
                             .upsert(self._upsert)
                             .execute())
            if return_id_list:
                id_list.append(last_id)

        if return_id_list:
            return id_list
        else:
            return last_id

    async def execute(self):
        insert_with_loop = (
            self._is_multi_row_insert and
            self._query is None and
            self._returning is None and
            not self.database.insert_many)
        if insert_with_loop:
            return await self._insert_with_loop()

        if self._returning is not None and self._qr is None:
            return await self._execute_with_result_wrapper()
        elif self._qr is not None:
            return self._qr
        else:
            cursor = await self._execute()
            if not self._is_multi_row_insert:
                if self.database.insert_returning:
                    pk_row = await cursor.fetchone()
                    meta = self.model_class._meta
                    clean_data = [
                        field.python_value(column)
                        for field, column
                        in zip(meta.get_primary_key_fields(), pk_row)]
                    if self.model_class._meta.composite_key:
                        return clean_data
                    return clean_data[0]
                return self.database.last_insert_id(cursor, self.model_class)
            elif self._return_id_list:
                return map(operator.itemgetter(0), await cursor.fetchall())
            else:
                return True


class AioDeleteQuery(_AioWriteQuery, DeleteQuery):

    async def execute(self):
        if self._returning is not None and self._qr is None:
            return await self._execute_with_result_wrapper()
        elif self._qr is not None:
            return self._qr
        else:
            return self.database.rows_affected(await self._execute())



import aiopg
import psycopg2
from peewee import basestring
from playhouse.postgres_ext import PostgresqlExtDatabase
from psycopg2 import ProgrammingError
from .database import AioDatabase


class AioPostgreSQLDatabase(AioDatabase, PostgresqlExtDatabase):

    async def _connect(self, database, loop=None, **kwargs):

        conn_kwargs = {
        }
        conn_kwargs.update(kwargs)
        ud = " ".join([str(i)+"="+str(j) for i,j in conn_kwargs.items()])
        dsn = 'dbname={} '.format(database)+ud
        return await aiopg.create_pool(dsn)


    async def _close(self, conn):
        self._conn_pool.release(conn)

    async def cancel(self,con,timeout=None):
        con.cancel(timeout=timeout)

    async def execute_sql(self, sql, params=None, require_commit=True):
        with self.exception_wrapper:
            conn = await self.get_conn()
            try:
                cursor = await conn.cursor()
                await cursor.execute(sql, params or ())
                # TODO: MIGHT CLOSE THE CURSOR FROM RESULT WRAPPER...
            except Exception:
                await self.cancel(conn)
                raise

            finally:
                await self._close(conn)

        return cursor

    async def get_tables(self, schema=None):
        cursor = await self.execute_sql("SELECT table_name FROM information_schema.tables WHERE table_schema = 'public';")
        return [row for row, in await cursor.fetchall()]

    async def get_indexes(self, table, schema=None):
        cursor = await self.execute_sql("SELECT * FROM pg_indexes WHERE tablename = '{}';".format(table))
        unique = set()
        indexes = {}
        for row in cursor.fetchall():
            if not row[1]:
                unique.add(row[2])
            indexes.setdefault(row[2], [])
            indexes[row[2]].append(row[4])
        return [IndexMetadata(name, None, indexes[name], name in unique, table)
                for name in indexes]

    async def get_columns(self, table, schema=None):
        sql = """
            SELECT column_name, is_nullable, data_type
            FROM information_schema.columns
            WHERE table_name = %s AND table_schema = DATABASE()"""
        cursor = await self.execute_sql(sql, (table,))
        pks = set(self.get_primary_keys(table))
        return [ColumnMetadata(name, dt, null == 'YES', name in pks, table)
                for name, null, dt in await cursor.fetchall()]

    async def get_primary_keys(self, table, schema=None):
        cursor = await self.execute_sql('SHOW INDEX FROM `%s`' % table)
        return [row[4] for row in await cursor.fetchall()
                if row[2] == 'PRIMARY']

    async def get_foreign_keys(self, table, schema=None):
        query = """
            SELECT column_name, referenced_table_name, referenced_column_name
            FROM information_schema.key_column_usage
            WHERE table_name = %s
                AND table_schema = DATABASE()
                AND referenced_table_name IS NOT NULL
                AND referenced_column_name IS NOT NULL"""
        cursor = await self.execute_sql(query, (table,))
        return [
            ForeignKeyMetadata(column, dest_table, dest_column, table)
            for column, dest_table, dest_column in await cursor.fetchall()]

    async def commit(self):
        raise psycopg2.ProgrammingError

    async def rollback(self):
        raise psycopg2.ProgrammingError


    # TODO
    def get_binary_type(self):
        return psycopg2.Binary



from unittest import TestCase
from aioorm.uri_parser import parser
from aioorm.error import InvalidURI
import platform
import getpass

class MysqlURITest(TestCase):

    def test_mysql(self):
        uri = "mysql:///test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'root',
            password = '',
            host = 'localhost',
            port = 3306,
            database = 'test'),
            parser(uri))

    def test_mysql_host(self):
        uri = "mysql://192.168.1.23/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'root',
            password = '',
            host = '192.168.1.23',
            port = 3306,
            database = 'test'),
            parser(uri))

    def test_mysql_port(self):
        uri = "mysql://:123/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'root',
            password = '',
            host = 'localhost',
            port = 123,
            database = 'test'),
            parser(uri))

    def test_mysql_user(self):
        uri = "mysql://user@/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'user',
            password = '',
            host = 'localhost',
            port = 3306,
            database = 'test'),
            parser(uri))

    def test_mysql_password(self):
        uri = "mysql://:password@/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'root',
            password = 'password',
            host = 'localhost',
            port = 3306,
            database = 'test'),
            parser(uri))

    def test_mysql_host_port(self):
        uri = "mysql://192.168.1.23:123/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'root',
            password = '',
            host = '192.168.1.23',
            port = 123,
            database = 'test'),
            parser(uri))

    def test_mysql_user_pass(self):
        uri = "mysql://user:password@/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'user',
            password = 'password',
            host = 'localhost',
            port = 3306,
            database = 'test'),
            parser(uri))

    def test_mysql_user_host(self):
        uri = "mysql://user@192.168.1.23/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'user',
            password = '',
            host = '192.168.1.23',
            port = 3306,
            database = 'test'),
            parser(uri))

    def test_mysql_user_port(self):
        uri = "mysql://user@:123/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'user',
            password = '',
            host = 'localhost',
            port = 123,
            database = 'test'),
            parser(uri))

    def test_mysql_host__pass(self):
        uri = "mysql://:password@192.168.1.23/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'root',
            password = 'password',
            host = '192.168.1.23',
            port = 3306,
            database = 'test'),
            parser(uri))

    def test_mysql_port_pass(self):
        uri = "mysql://:password@:123/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'root',
            password = 'password',
            host = 'localhost',
            port = 123,
            database = 'test'),
            parser(uri))

    def test_mysql_port_user_pass(self):
        uri = "mysql://user:password@:123/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'user',
            password = 'password',
            host = 'localhost',
            port = 123,
            database = 'test'),
            parser(uri))

    def test_mysql_host_user_pass(self):
        uri = "mysql://user:password@192.168.1.23/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'user',
            password = 'password',
            host = '192.168.1.23',
            port = 3306,
            database = 'test'),
            parser(uri))

    def test_mysql_host_port_pass(self):
        uri = "mysql://:password@192.168.1.23:123/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'root',
            password = 'password',
            host = '192.168.1.23',
            port = 123,
            database = 'test'),
            parser(uri))

    def test_mysql_host_port_user(self):
        uri = "mysql://user@192.168.1.23:123/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'user',
            password = '',
            host = '192.168.1.23',
            port = 123,
            database = 'test'),
            parser(uri))

    def test_mysql_host_port_user_pass(self):
        uri = "mysql://user:password@192.168.1.23:123/test"
        self.assertDictEqual(dict(scheme="mysql",
            username = 'user',
            password = 'password',
            host = '192.168.1.23',
            port = 123,
            database = 'test'),
            parser(uri))

class PostgresqlURITest(TestCase):

    def setUp(self):
        if platform.system() == 'Darwin':
            self.user = getpass.getuser()
            self.password = ''
        elif platform.system() == 'windows':
            self.user = 'postgres'
            self.password = 'postgres'
        else:
            self.user = 'postgres'
            self.password = ''

        self.host='localhost'
        self.port=5432

    def test_postgresql(self):
        uri = "postgresql:///test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = self.user,
            password = self.password,
            host = self.host,
            port = self.port,
            database = 'test'),
            parser(uri))

    def test_postgresql_host(self):
        uri = "postgresql://192.168.1.23/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = self.user,
            password = self.password,
            host = "192.168.1.23",
            port = self.port,
            database = 'test'),
            parser(uri))

    def test_postgresql_port(self):
        uri = "postgresql://:123/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = self.user,
            password = self.password,
            host = self.host,
            port = 123,
            database = 'test'),
            parser(uri))

    def test_postgresql_user(self):
        uri = "postgresql://user@/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = 'user',
            password = self.password,
            host = self.host,
            port = self.port,
            database = 'test'),
            parser(uri))

    def test_postgresql_password(self):
        uri = "postgresql://:password@/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = self.user,
            password = 'password',
            host = self.host,
            port = self.port,
            database = 'test'),
            parser(uri))

    def test_postgresql_host_port(self):
        uri = "postgresql://192.168.1.23:123/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = self.user,
            password = self.password,
            host = '192.168.1.23',
            port = 123,
            database = 'test'),
            parser(uri))

    def test_postgresql_user_pass(self):
        uri = "postgresql://user:password@/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = 'user',
            password = 'password',
            host = self.host,
            port = self.port,
            database = 'test'),
            parser(uri))

    def test_postgresql_user_host(self):
        uri = "postgresql://user@192.168.1.23/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = 'user',
            password = self.password,
            host = '192.168.1.23',
            port = self.port,
            database = 'test'),
            parser(uri))

    def test_postgresql_user_port(self):
        uri = "postgresql://user@:123/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = 'user',
            password = self.password,
            host = self.host,
            port = 123,
            database = 'test'),
            parser(uri))

    def test_postgresql_host__pass(self):
        uri = "postgresql://:password@192.168.1.23/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = self.user,
            password = 'password',
            host = '192.168.1.23',
            port = self.port,
            database = 'test'),
            parser(uri))

    def test_postgresql_port_pass(self):
        uri = "postgresql://:password@:123/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = self.user,
            password = 'password',
            host = self.host,
            port = 123,
            database = 'test'),
            parser(uri))

    def test_postgresql_port_user_pass(self):
        uri = "postgresql://user:password@:123/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = 'user',
            password = 'password',
            host = self.host,
            port = 123,
            database = 'test'),
            parser(uri))


    def test_postgresql_host_user_pass(self):
        uri = "postgresql://user:password@192.168.1.23/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = 'user',
            password = 'password',
            host = '192.168.1.23',
            port = self.port,
            database = 'test'),
            parser(uri))

    def test_postgresql_host_port_pass(self):
        uri = "postgresql://:password@192.168.1.23:123/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = self.user,
            password = 'password',
            host = '192.168.1.23',
            port = 123,
            database = 'test'),
            parser(uri))

    def test_postgresql_host_port_user(self):
        uri = "postgresql://user@192.168.1.23:123/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = 'user',
            password = self.password,
            host = '192.168.1.23',
            port = 123,
            database = 'test'),
            parser(uri))

    def test_postgresql_host_port_user_pass(self):
        uri = "postgresql://user:password@192.168.1.23:123/test"
        self.assertDictEqual(dict(scheme="postgresql",
            username = 'user',
            password = 'password',
            host = '192.168.1.23',
            port = 123,
            database = 'test'),
            parser(uri))

class ParserErrorTest(TestCase):

    def test_unknow_scheme(self):
        with self.assertRaisesRegex(InvalidURI,r'unknow database') as a:
            parser('sqlit:///test')

    def test_without_datebase(self):
        with self.assertRaisesRegex(InvalidURI,r"need to point out the db's name") as a:
            parser('mysql://')

    def test_without_scheme(self):
        with self.assertRaisesRegex(InvalidURI,r'must have a database uri') as a:
            parser(':///test')

    def test_not_a_scheme(self):
        with self.assertRaisesRegex(InvalidURI,r'must have a database uri') as a:
            parser(':adf/test')



from functools import partial
from collections import Collection, Mapping
import asyncio
import aiofiles
import os
from peewee import Func
from peewee import Field

class csv_writer:
    def __init__(self,fh,fieldnames=None,delimiter=','):
        self.fh = fh
        self.delimiter = delimiter
        if fieldnames:
            self.fieldnames = list(fieldnames)
        else:
            self.fieldnames = None

    async def writerow(self,row):
        if isinstance(row,Collection):
            row = (self.delimiter).join([str(i) for i in row])
        elif isinstance(row,Mapping):
            if not self.fieldnames:
                raise AttributeError("do not have header")
            else:
                row_ = [row.get(i,"") for i in self.fieldnames]
                row = (self.delimiter).join(row_)
        else:
            raise AttributeError("unsupport row type")
        try:
            await self.fh.write(row+os.linesep)
        except TypeError as te:
            #self.fh.write(row+os.linesep)
            pass
        except:
            raise

    async def writeheader(self,fieldnames=None):
        if fieldnames:
            self.fieldnames = list(fieldnames)
        if self.fieldnames:
            row = (self.delimiter).join(self.fieldnames)
            try:
                await self.fh.write(row+os.linesep)
            except TypeError as te:
                #self.fh.write(row+os.linesep)
                pass
            except:
                raise


async def aiodump_csv(query, file_or_name, loop= None,include_header=True, close_file=False,
             append=True, csv_writer=csv_writer):

    if isinstance(file_or_name, str):
        close_file=True
        #fh = await aiofiles.open(file_or_name, append and 'a' or 'w')
        fh = await aiofiles.open(file_or_name, 'w')
        writer = csv_writer(fh)
    else:
        fh = file_or_name
        if append:
            try:
                await fh.seek(0, 2)
            except TypeError as te:
                fh.seek(0, 2)
            except:
                raise

        writer = csv_writer(fh)

    if include_header:
        header = []
        for idx, node in enumerate(query._select):
            if node._alias:
                header.append(node._alias)
            elif isinstance(node, (Field, Func)):
                header.append(node.name)
            else:
                header.append('col_%s' % idx)
        await writer.writeheader(header)

    for row in (await query.tuples()):
        await writer.writerow(row)
    if close_file:
        try:
            await fh.close()
        except TypeError as te:
            fh.close()
        except:
            raise

    return fh



import re
from playhouse.fields import (ManyToManyField, ManyToManyQuery,
                              ManyToManyFieldDescriptor, DeferredThroughModel)
from peewee import Proxy, ForeignKeyField, SelectQuery, Model

from .model import AioModel
from .query import AioSelectQuery


class AioManyToManyField(ManyToManyField):

    def _get_descriptor(self):
        return AioManyToManyFieldDescriptor(self)

    def add_to_class(self, model_class, name):
        if isinstance(self._through_model, Proxy):
            def callback(through_model):
                self._through_model = through_model
                self.add_to_class(model_class, name)
            self._through_model.attach_callback(callback)
            return
        elif isinstance(self._through_model, DeferredThroughModel):
            self._through_model.set_field(model_class, self, name)
            return

        self.name = name
        self.model_class = model_class
        if not self.verbose_name:
            self.verbose_name = re.sub('_+', ' ', name).title()
        setattr(model_class, name, self._get_descriptor())

        if not self._is_backref:
            backref = AioManyToManyField(
                self.model_class,
                through_model=self._through_model,
                _is_backref=True)
            related_name = self._related_name or model_class._meta.name + 's'
            backref.add_to_class(self.rel_model, related_name)

    def get_through_model(self):
        if not self._through_model:
            lhs, rhs = self.get_models()
            tables = [model._meta.db_table for model in (lhs, rhs)]

            class Meta:
                database = self.model_class._meta.database
                db_table = '%s_%s_through' % tuple(tables)
                indexes = (
                    ((lhs._meta.name, rhs._meta.name),
                     True),)
                validate_backrefs = False

            attrs = {
                lhs._meta.name: ForeignKeyField(rel_model=lhs),
                rhs._meta.name: ForeignKeyField(rel_model=rhs)}
            attrs['Meta'] = Meta

            self._through_model = type(
                '%s%sThrough' % (lhs.__name__, rhs.__name__),
                (AioModel,),
                attrs)

        return self._through_model


class AioManyToManyFieldDescriptor(ManyToManyFieldDescriptor):

    def __get__(self, instance, instance_type=None):
        if instance is not None:
            return (AioManyToManyQuery(instance, self, self.rel_model)
                    .select()
                    .join(self.through_model)
                    .join(self.model_class)
                    .where(self.src_fk == instance))
        return self.field

    def __set__(self, instance, value):
        raise NotImplementedError()
        # query = self.__get__(instance)
        # query.add(value, clear_existing=True)


class AioManyToManyQuery(AioSelectQuery, ManyToManyQuery):

    # TODO
    def _id_list(self, model_or_id_list):
        if isinstance(model_or_id_list[0], Model):
            return [obj.get_id() for obj in model_or_id_list]
        return model_or_id_list

    async def set(self, value):
        await self.add(value, clear_existing=True)

    async def add(self, value, clear_existing=False):
        if clear_existing:
            await self.clear()

        fd = self._field_descriptor
        if isinstance(value, SelectQuery):
            query = value.select(
                SQL(str(self._instance.get_id())),
                fd.rel_model._meta.primary_key)
            await fd.through_model.insert_from(
                fields=[fd.src_fk, fd.dest_fk],
                query=query).execute()
        else:
            if not isinstance(value, (list, tuple)):
                value = [value]
            if not value:
                return
            inserts = [{
                fd.src_fk.name: self._instance.get_id(),
                fd.dest_fk.name: rel_id}
                for rel_id in self._id_list(value)]
            await fd.through_model.insert_many(inserts).execute()

    async def remove(self, value):
        fd = self._field_descriptor
        if isinstance(value, SelectQuery):
            subquery = value.select(value.model_class._meta.primary_key)
            return await (fd.through_model
                            .delete()
                            .where(
                                (fd.dest_fk << subquery) &
                                (fd.src_fk == self._instance.get_id()))
                            .execute())
        else:
            if not isinstance(value, (list, tuple)):
                value = [value]
            if not value:
                return
            return await (fd.through_model
                            .delete()
                            .where(
                                (fd.dest_fk << self._id_list(value)) &
                                (fd.src_fk == self._instance.get_id()))
                            .execute())

    async def clear(self):
        return await (self._field_descriptor.through_model
                      .delete()
                      .where(self._field_descriptor.src_fk == self._instance)
                      .execute())



from aitertools import alist

from aioorm import AioModel, AioMySQLDatabase, AioManyToManyField,AioPostgreSQLDatabase
from peewee import (ForeignKeyField, IntegerField, CharField,Proxy,
                    DateTimeField, TextField, PrimaryKeyField)

from unittest import TestCase
import asyncio

db = Proxy()

class User(AioModel):
    username = CharField(unique=True)

    class Meta:
        database = db


class Note(AioModel):
    text = TextField()
    users = AioManyToManyField(User)

    class Meta:
        database = db


NoteUserThrough = Note.users.get_through_model()

async def create_users_notes():
    usernames = ['charlie', 'huey', 'mickey', 'zaizee']
    n_notes = 5
    for username in usernames:
        await User.create(username=username)
    for i in range(n_notes):
        await Note.create(text='note-%s' % (i + 1))

    # create relationship
    user_to_note = {'charlie': [1, 2],
                    'huey': [2, 3],
                    'mickey': [3, 4],
                    'zaizee': [4, 5]}
    for username, notes in user_to_note.items():
        user = await User.get(User.username == username)
        for note in notes:
            note = await Note.get(Note.text == 'note-%s' % note)
            await NoteUserThrough.create(note=note, user=user)


async def assert_notes(query, expected):
    notes = [note.text async for note in query]
    assert sorted(notes) == ['note-%s' % i for i in sorted(expected)]


async def assert_users(query, expected):
    usernames = [user.username async for user in query]
    assert sorted(usernames) == sorted(expected)


async def set_values(loop):
    await db.connect(loop)
    await db.create_tables([User, Note, NoteUserThrough], safe=True)
    await create_users_notes()

    charlie = await User.get(User.username == 'charlie')
    huey = await User.get(User.username == 'huey')
    n1, n2, n3, n4, n5 = await alist(Note.select().order_by(Note.text))

    await assert_notes(charlie.notes, [1, 2])
    await assert_users(n1.users, ['charlie'])

    await charlie.notes.set([n2, n3])
    await assert_notes(charlie.notes, [2, 3])
    await assert_users(n1.users, [])
    await assert_users(n2.users, ['charlie', 'huey'])
    await assert_users(n3.users, ['charlie', 'huey', 'mickey'])

    await db.drop_tables([User, Note, NoteUserThrough], safe=True)
    await db.close()

class ManyToManyMysqlTest(TestCase):
        @classmethod
        def setUpClass(cls, *args, **kwargs):
            """Configure database managers, create test tables.
            """
            database = AioMySQLDatabase('test', host='localhost', port=3306,
                       user='root', password='hsz881224')
            db.initialize(database)
            cls.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(cls.loop)
        @classmethod
        def tearDownClass(cls, *args, **kwargs):
            """Remove all test tables and close connections.
            """
            cls.loop.close()

        def test_mtm(self):
            self.loop.run_until_complete(set_values(self.loop))

class ManyToManyPostgreSQLTest(TestCase):
        @classmethod
        def setUpClass(cls, *args, **kwargs):
            """Configure database managers, create test tables.
            """
            database = AioPostgreSQLDatabase('test_ext', host='localhost', port=5432,
                      user='huangsizhe', password='')
            db.initialize(database)
            cls.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(cls.loop)
        @classmethod
        def tearDownClass(cls, *args, **kwargs):
            """Remove all test tables and close connections.
            """
            cls.loop.close()

        def test_mtm(self):
            self.loop.run_until_complete(set_values(self.loop))



import aiomysql
import pymysql
from peewee import MySQLDatabase
from .context import _aio_atomic, aio_transaction, aio_savepoint
from .database import AioDatabase


class AioMySQLDatabase(AioDatabase, MySQLDatabase):

    async def close(self):
        if self.deferred:
            raise Exception('Error, database not properly initialized '
                            'before closing connection')
        with self.exception_wrapper:
            if not self._closed and self._conn_pool:
                #关闭的时候先关闭自动链接
                await self.close_engine()
                self._conn_pool.close()
                self._closed = True
                await self._conn_pool.wait_closed()

    async def connect(self, loop=None):
        if self.deferred:
            raise OperationalError('Database has not been initialized')
        if not self._closed:
            raise OperationalError('Connection already open')
        self._conn_pool = await self._create_connection(loop=loop)
        self._closed = False
        # 启动自动链接
        await self.init_engine(loop=loop)

        with self.exception_wrapper:
            self.initialize_connection(self._conn_pool)

    async def init_engine(self, loop=None):
        # create engine
  
        self._auto_task = loop.create_task(keep_engine())

    async def close_engine(self):
        self._auto_task.cancel()

        # close engine

    async def keep_engine(self):
        while True:
            async with self._conn_pool.engine.acquire() as conn:
                await conn.connection.ping()

            await asyncio.sleep(60)

    def set_autocommit(self, autocommit):
        self.autocommit = autocommit

    def get_autocommit(self):
        if self.autocommit is None:
            self.set_autocommit(self.autocommit)
        return self.autocommit

    def transaction(self, transaction_type=None):
        return aio_transaction(self, transaction_type)
    commit_on_success = property(transaction)

    def savepoint(self, sid=None):
        if not self.savepoints:
            raise NotImplementedError
        return aio_savepoint(self, sid)

    def atomic(self, transaction_type=None):
        return _aio_atomic(self, transaction_type)

    async def commit(self):
        with self.exception_wrapper:
            await self.get_conn().commit()

    async def rollback(self):
        with self.exception_wrapper:
            await self.get_conn().rollback()

    async def _connect(self, database, loop=None, **kwargs):
        # if not mysql:
        #     raise ImproperlyConfigured('MySQLdb or PyMySQL must be installed.')
        conn_kwargs = {
            'charset': 'utf8',
            'use_unicode': True,
        }
        conn_kwargs.update(kwargs)
        return await aiomysql.create_pool(db=database, loop=loop,**conn_kwargs)

    async def get_tables(self, schema=None):
        cursor = await self.execute_sql('SHOW TABLES')
        return [row for row, in await cursor.fetchall()]

    async def get_indexes(self, table, schema=None):
        cursor = await self.execute_sql('SHOW INDEX FROM `%s`' % table)
        unique = set()
        indexes = {}
        for row in cursor.fetchall():
            if not row[1]:
                unique.add(row[2])
            indexes.setdefault(row[2], [])
            indexes[row[2]].append(row[4])
        return [IndexMetadata(name, None, indexes[name], name in unique, table)
                for name in indexes]

    async def get_columns(self, table, schema=None):
        sql = """
            SELECT column_name, is_nullable, data_type
            FROM information_schema.columns
            WHERE table_name = %s AND table_schema = DATABASE()"""
        cursor = await self.execute_sql(sql, (table,))
        pks = set(self.get_primary_keys(table))
        return [ColumnMetadata(name, dt, null == 'YES', name in pks, table)
                for name, null, dt in await cursor.fetchall()]

    async def get_primary_keys(self, table, schema=None):
        cursor = await self.execute_sql('SHOW INDEX FROM `%s`' % table)
        return [row[4] for row in await cursor.fetchall()
                if row[2] == 'PRIMARY']

    async def get_foreign_keys(self, table, schema=None):
        query = """
            SELECT column_name, referenced_table_name, referenced_column_name
            FROM information_schema.key_column_usage
            WHERE table_name = %s
                AND table_schema = DATABASE()
                AND referenced_table_name IS NOT NULL
                AND referenced_column_name IS NOT NULL"""
        cursor = await self.execute_sql(query, (table,))
        return [
            ForeignKeyMetadata(column, dest_table, dest_column, table)
            for column, dest_table, dest_column in await cursor.fetchall()]

    # TODO
    def get_binary_type(self):
        return pymysql.Binary



from unittest import TestCase,expectedFailure
from aioorm.utils import AioDbFactory
from aioorm.postgresql import AioPostgreSQLDatabase
from aioorm.mysql import AioMySQLDatabase

class  AioDbFactoryTest(TestCase):

    def test_mysql(self):
        assert isinstance(AioDbFactory("mysql:///test"), AioMySQLDatabase)

    def test_Postgre(self):
        assert isinstance(AioDbFactory("postgresql:///test"), AioPostgreSQLDatabase)

    @expectedFailure
    def test_Failure(self):
        AioDbFactory("sqlit://")




from aitertools import alist
from datetime import datetime
from aioorm import AioModel, AioMySQLDatabase,AioPostgreSQLDatabase
from peewee import (ForeignKeyField, IntegerField, CharField,
                    DateTimeField, TextField, PrimaryKeyField,Proxy,
                    SQL)

from unittest import TestCase
import asyncio

db = Proxy()

class Count(AioModel):
    id = IntegerField(primary_key=True)
    data = CharField(max_length=64, null=False)
    count = IntegerField(default=0)
    created_at = DateTimeField()
    updated_at = DateTimeField()

    class Meta:
        database = db

class User(AioModel):
    username = CharField()

    class Meta:
        database = db

    def prepared(self):
        self.foo = self.username

    @classmethod
    async def create_users(cls, n):
        for i in range(n):
            await cls.create(username='u%d' % (i + 1))

class Blog(AioModel):
    user = ForeignKeyField(User)
    title = CharField(max_length=25)
    content = TextField(default='')
    pub_date = DateTimeField(null=True)
    pk = PrimaryKeyField()

    class Meta:
        database = db

    def prepared(self):
        self.foo = self.title

async def create_users_blogs(n=10, nb=5):
    for i in range(n):
        u = await User.create(username='u%d' % i)
        for j in range(nb):
            b = await Blog.create(title='b-%d-%d' % (i, j),
                                  content=str(j), user=u)

async def test_table_creation(loop):
    await db.connect(loop)

    assert await Count.table_exists() is False

    await Count.create_table()
    assert await Count.table_exists() is True

    await Count.drop_table()
    assert await Count.table_exists() is False

    await db.close()

async def test_crud(loop):
    await db.connect(loop)
    await Count.create_table(fail_silently=True)

    now = datetime.now()
    await Count.create(id=1, data='test', created_at=now, updated_at=now)
    await Count.create(id=2, data='test', created_at=now, updated_at=now)

    # TODO
    await Count.drop_table()
    await db.close()

async def test_query(loop):
    await db.connect(loop)
    await Count.create_table(fail_silently=True)

    now = datetime.now()
    await Count.create(id=1, data='test', created_at=now, updated_at=now)
    await Count.create(id=2, data='test', created_at=now, updated_at=now)

    c = await Count.select().count()
    assert c == 2, ''

    data = [i async for i in Count.select()]
    assert len(data) == 2, ''

    data = [i async for i in Count.select().where(Count.id > 0)]
    assert len(data) == 2, ''

    data = [i async for i in Count.select().limit(1)]
    assert len(data) == 1, ''

    data = [i async for i in Count.select().order_by(Count.id.desc())]
    assert data[0].id == 2

    record = await Count.select().order_by(Count.id.desc()).first()
    assert record.id == 2

    # TODO
    # t = await Count.select().order_by(Count.id.desc()).first()
    # await t.delete_instance()

    # await Count.update(data='12345')
    # t = await Count.select().order_by(Count.id.desc()).first()
    # assert t.data == '12345', ''

    await Count.drop_table()
    await db.close()

async def test_select(loop):
    await db.connect(loop)
    await db.create_tables([User, Blog], safe=True)
    await create_users_blogs()

    users = (User.select()
                 .where(User.username << ['u0', 'u5'])
                 .order_by(User.username))
    assert [u.username async for u in users] == ['u0', 'u5']

    blogs = Blog.select().join(User).where(
        (User.username << ['u0', 'u3']) &
        (Blog.content == '4')
    ).order_by(Blog.title)
    assert [b.title async for b in blogs] == ['b-0-4', 'b-3-4']

    users = User.select().paginate(2, 3)
    assert [u.username async for u in users] == ['u3', 'u4', 'u5']

    await db.drop_tables([User, Blog], safe=True)
    await db.close()

async def test_select_all(loop):
    await db.connect(loop)
    await db.create_tables([User, Blog], safe=True)
    await create_users_blogs(2, 2)

    all_cols = SQL('*')
    query = Blog.select(all_cols)
    blogs = [blog async for blog in query.order_by(Blog.pk)]
    assert [b.title for b in blogs] == ['b-0-0', 'b-0-1', 'b-1-0', 'b-1-1']

    assert [(await b.user).username for b in blogs] == ['u0', 'u0', 'u1', 'u1']

    # TODO: await blogs[0] fails with cannot await again
    # must be a caching issue

    await db.drop_tables([User, Blog], safe=True)
    await db.close()

async def test_select_all_fetchall(loop):
    from aitertools import aiter
    await db.connect(loop)
    await db.create_tables([User, Blog], safe=True)
    await create_users_blogs(2, 2)

    all_cols = SQL('*')

    query0 = Blog.select(all_cols).order_by(Blog.pk)
    blogs0 = await alist(query0)

    query1 = Blog.select(all_cols).order_by(Blog.pk)
    blogs1 = await query1

    assert isinstance(blogs0, list)
    assert isinstance(blogs1, list)
    assert [b.title for b in blogs0] == ['b-0-0', 'b-0-1', 'b-1-0', 'b-1-1']
    assert [b.title for b in blogs1] == ['b-0-0', 'b-0-1', 'b-1-0', 'b-1-1']
    assert blogs0 == blogs1

    await db.drop_tables([User, Blog], safe=True)
    await db.close()

async def test_insert(loop):
    await db.connect(loop)
    await db.create_tables([User], safe=True)

    iq = User.insert(username='u1')
    assert await User.select().count() == 0

    uid = await iq.execute()
    assert uid > 0

    assert await User.select().count() == 1
    u = await User.get(User.id==uid)
    u.username == 'u1'

    try:
        await User.insert(doesnotexist='invalid')
    except Exception as e:
        assert isinstance(e,KeyError)
    else:
        assert 0



    await db.drop_tables([User], safe=True)
    await db.close()

async def test_insert_many(loop):
    await db.connect(loop)
    await db.create_tables([User], safe=True)

    iq = User.insert_many([
        {'username': 'u1'},
        {'username': 'u2'},
        {'username': 'u3'},
        {'username': 'u4'}])
    assert await iq.execute() is True

    await User.select().count() == 4

    sq = User.select(User.username).order_by(User.username)
    assert [u.username async for u in sq] == ['u1', 'u2', 'u3', 'u4']

    iq = User.insert_many([{'username': 'u5'}])
    await iq.execute() is True
    await User.select().count() == 5

    iq = await User.insert_many([
        {User.username: 'u6'},
        {User.username: 'u7'},
        {'username': 'u8'}])

    sq = User.select(User.username).order_by(User.username)
    assert [u.username async for u in sq] == ['u1', 'u2', 'u3', 'u4',
                                              'u5', 'u6', 'u7', 'u8']

    await db.drop_tables([User], safe=True)
    await db.close()

async def test_delete(loop):
    await db.connect(loop)
    await db.create_tables([User], safe=True)

    await User.create_users(5)
    dq = User.delete().where(User.username << ['u1', 'u2', 'u3'])
    assert await User.select().count() == 5
    nr = await dq
    assert nr == 3
    assert [u.username async for u in User.select()] == ['u4', 'u5']

    await db.drop_tables([User], safe=True)
    await db.close()

async def test_related_name(loop):
    await db.connect(loop)
    await db.create_tables([User, Blog], safe=True)

    u1 = await User.create(username='u1')
    u2 = await User.create(username='u2')
    b11 = await Blog.create(user=u1, title='b11')
    b12 = await Blog.create(user=u1, title='b12')
    b2 = await Blog.create(user=u2, title='b2')

    assert [b.title async for b in u1.blog_set.order_by(Blog.title)] == ['b11', 'b12']
    assert [b.title async for b in u2.blog_set.order_by(Blog.title)] == ['b2']

    await db.drop_tables([User, Blog], safe=True)
    await db.close()

async def test_saving(loop):
    await db.connect(loop)
    await db.create_tables([User], safe=True)

    assert await User.select().count() == 0

    u = User(username='u1')
    assert await u.save() == 1
    u.username = 'u2'
    assert await u.save() == 1

    assert await User.select().count() == 1

    assert await u.delete_instance() == 1
    assert await u.save() == 0

    await db.drop_tables([User], safe=True)
    await db.close()

class ModelMysqlTest(TestCase):
        @classmethod
        def setUpClass(cls, *args, **kwargs):
            """Configure database managers, create test tables.
            """
            database = AioMySQLDatabase('test', host='localhost', port=3306,
                       user='root', password='hsz881224')
            db.initialize(database)
            cls.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(cls.loop)
        @classmethod
        def tearDownClass(cls, *args, **kwargs):
            """Remove all test tables and close connections.
            """
            cls.loop.close()

        def test_table_creation(self):
            self.loop.run_until_complete(test_table_creation(self.loop))

        def test_crud(self):
            self.loop.run_until_complete(test_crud(self.loop))

        def test_query(self):
            self.loop.run_until_complete(test_query(self.loop))

        def test_select(self):
            self.loop.run_until_complete(test_select(self.loop))

        def test_select_all(self):
            self.loop.run_until_complete(test_select_all(self.loop))

        def test_select_all_fetchall(self):
            self.loop.run_until_complete(test_select_all_fetchall(self.loop))

        def test_insert(self):
            self.loop.run_until_complete(test_insert(self.loop))

        def test_insert_many(self):
            self.loop.run_until_complete(test_insert_many(self.loop))

        def test_delete(self):
            self.loop.run_until_complete(test_delete(self.loop))

        def test_related_name(self):
            self.loop.run_until_complete(test_related_name(self.loop))

        def test_saving(self):
            self.loop.run_until_complete(test_saving(self.loop))


class ModelPostgreSQLTest(TestCase):
        @classmethod
        def setUpClass(cls, *args, **kwargs):
            """Configure database managers, create test tables.
            """
            database = AioPostgreSQLDatabase('test_ext', host='localhost', port=5432,
                      user='huangsizhe', password='')
            db.initialize(database)
            cls.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(cls.loop)
        @classmethod
        def tearDownClass(cls, *args, **kwargs):
            """Remove all test tables and close connections.
            """
            cls.loop.close()

        def test_table_creation(self):
            self.loop.run_until_complete(test_table_creation(self.loop))

        def test_crud(self):
            self.loop.run_until_complete(test_crud(self.loop))

        def test_query(self):
            self.loop.run_until_complete(test_query(self.loop))

        def test_select(self):
            self.loop.run_until_complete(test_select(self.loop))

        def test_select_all(self):
            self.loop.run_until_complete(test_select_all(self.loop))

        def test_select_all_fetchall(self):
            self.loop.run_until_complete(test_select_all_fetchall(self.loop))

        def test_insert(self):
            self.loop.run_until_complete(test_insert(self.loop))

        def test_insert_many(self):
            self.loop.run_until_complete(test_insert_many(self.loop))

        def test_delete(self):
            self.loop.run_until_complete(test_delete(self.loop))

        def test_related_name(self):
            self.loop.run_until_complete(test_related_name(self.loop))

        def test_saving(self):
            self.loop.run_until_complete(test_saving(self.loop))



from ..postgresql import AioPostgreSQLDatabase
from ..mysql import AioMySQLDatabase
from ..uri_parser import parser

from typing import TypeVar
AnyDb = TypeVar('AnyDb', AioMySQLDatabase, AioPostgreSQLDatabase)

DBS = {
    'mysql':AioMySQLDatabase,
    'postgresql':AioPostgreSQLDatabase
}



def AioDbFactory(uri:str)->AnyDb:
    info = parser(uri)
    db = DBS.get(info.get("scheme"))(
        info.get('database'),
        host=info.get("host"),
        port=info.get("port"),
        user=info.get("username"),
        password=info.get("password"))
    print(db)
    return db



import aiofiles
from io import StringIO
from peewee import *
from peewee import Database
import csv
import _csv
class Reader:

    def __init__(self,fh,delimiter=',',**kws):
        self.fh = fh
        self.delimiter = delimiter
        self.kws=dict(**kws)

    def __aiter__(self):
        return self

    async def __anext__(self):
        line = await self.fh.readline()
        if line:
            row = line.strip().split(self.delimiter)
            return row
        else:
            raise StopAsyncIteration


class Get_Reader:

    def __init__(self, file_or_name, **reader_kwargs):
        self.file_or_name=file_or_name
        self.reader_kwargs = dict(**reader_kwargs)
        self.is_file = False
        self.is_io = False
    async def __aenter__(self):

        if isinstance(self.file_or_name, str):
            self.fh = await aiofiles.open(self.file_or_name, 'r')
        elif isinstance(self.file_or_name, StringIO):
            self.fh = self.file_or_name
            self.fh.seek(0)
            self.is_io=True
        else:
            self.fh = self.file_or_name
            await self.fh.seek(0)
            self.is_file = True

        if self.is_io:
            reader = csv.reader(self.fh, **self.reader_kwargs)
        else:
            reader = Reader(self.fh, **self.reader_kwargs)
        return reader
    async def __aexit__(self, exc_type, exc, tb):
        if self.is_file:
            await self.fh.close()
        if self.is_io:
            self.fh.close()

class _CSVReader():

    def get_reader(self, file_or_name, **reader_kwargs):
        return Get_Reader(file_or_name, **reader_kwargs)

def convert_field(field_class, **field_kwargs):
    def decorator(fn):
        fn.field = lambda: field_class(**field_kwargs)
        return fn
    return decorator

class RowConverter(_CSVReader):
    """
    Simple introspection utility to convert a CSV file into a list of headers
    and column types.
    :param database: a peewee Database object.
    :param bool has_header: whether the first row of CSV is a header row.
    :param int sample_size: number of rows to introspect
    """
    date_formats = [
        '%Y-%m-%d',
        '%m/%d/%Y']

    datetime_formats = [
        '%Y-%m-%d %H:%M:%S',
        '%Y-%m-%d %H:%M:%S.%f']

    def __init__(self, database, has_header=True, sample_size=10):
        self.database = database
        self.has_header = has_header
        self.sample_size = sample_size

    def matches_date(self, value, formats):
        for fmt in formats:
            try:
                datetime.datetime.strptime(value, fmt)
            except ValueError:
                pass
            else:
                return True

    @convert_field(IntegerField, default=0)
    def is_integer(self, value):
        return value.isdigit()

    @convert_field(FloatField, default=0)
    def is_float(self, value):
        try:
            float(value)
        except (ValueError, TypeError):
            pass
        else:
            return True

    @convert_field(DateTimeField, null=True)
    def is_datetime(self, value):
        return self.matches_date(value, self.datetime_formats)

    @convert_field(DateField, null=True)
    def is_date(self, value):
        return self.matches_date(value, self.date_formats)

    @convert_field(BareField, default='')
    def default(self, value):
        return True

    async def extract_rows(self, file_or_name, **reader_kwargs):
        """
        Extract `self.sample_size` rows from the CSV file and analyze their
        data-types.
        :param str file_or_name: A string filename or a file handle.
        :param reader_kwargs: Arbitrary parameters to pass to the CSV reader.
        :returns: A 2-tuple containing a list of headers and list of rows
                  read from the CSV file.
        """
        rows = []
        rows_to_read = self.sample_size
        async with self.get_reader(file_or_name, **reader_kwargs) as reader:
            if self.has_header:
                rows_to_read += 1
            for i in range(self.sample_size):
                try:
                    row = await reader.__anext__()
                except AttributeError as te:
                    row = next(reader)
                except:
                    raise
                rows.append(row)

        if self.has_header:
            header, rows = rows[0], rows[1:]
        else:
            header = ['field_%d' % i for i in range(len(rows[0]))]
        return header, rows

    def get_checks(self):
        """Return a list of functions to use when testing values."""
        return [
            self.is_date,
            self.is_datetime,
            self.is_integer,
            self.is_float,
            self.default]

    def analyze(self, rows):
        """
        Analyze the given rows and try to determine the type of value stored.
        :param list rows: A list-of-lists containing one or more rows from a
                          csv file.
        :returns: A list of peewee Field objects for each column in the CSV.
        """
        transposed = zip(*rows)
        checks = self.get_checks()
        column_types = []
        for i, column in enumerate(transposed):
            # Remove any empty values.
            col_vals = [val for val in column if val != '']
            for check in checks:
                results = set(check(val) for val in col_vals)
                if all(results):
                    column_types.append(check.field())
                    break

        return column_types


class Loader(_CSVReader):
    """
    Load the contents of a CSV file into a database and return a model class
    suitable for working with the CSV data.
    :param db_or_model: a peewee Database instance or a Model class.
    :param file_or_name: the filename of the CSV file *or* a file handle.
    :param list fields: A list of peewee Field() instances appropriate to
        the values in the CSV file.
    :param list field_names: A list of names to use for the fields.
    :param bool has_header: Whether the first row of the CSV file is a header.
    :param int sample_size: Number of rows to introspect if fields are not
        defined.
    :param converter: A RowConverter instance to use.
    :param str db_table: Name of table to store data in (if not specified, the
        table name will be derived from the CSV filename).
    :param reader_kwargs: Arbitrary arguments to pass to the CSV reader.
    """
    def __init__(self, db_or_model, file_or_name, fields=None,
                 field_names=None, has_header=True, sample_size=10,
                 converter=None, db_table=None, pk_in_csv=False,
                 **reader_kwargs):
        self.file_or_name = file_or_name
        self.fields = fields
        self.field_names = field_names
        self.has_header = has_header
        self.sample_size = sample_size
        self.converter = converter
        self.reader_kwargs = reader_kwargs

        if isinstance(file_or_name, str):
            self.filename = file_or_name
        elif isinstance(file_or_name, StringIO):
            self.filename = 'data.csv'
        else:
            self.filename = file_or_name._file.name

        if isinstance(db_or_model, Database):
            self.database = db_or_model
            self.model = None
            self.db_table = (
                db_table or
                os.path.splitext(os.path.basename(self.filename))[0])
        else:
            self.model = db_or_model
            self.database = self.model._meta.database
            self.db_table = self.model._meta.db_table
            self.fields = self.model._meta.sorted_fields
            self.field_names = self.model._meta.sorted_field_names
            # If using an auto-incrementing primary key, ignore it unless we
            # are told the primary key is included in the CSV.
            if self.model._meta.auto_increment and not pk_in_csv:
                self.fields = self.fields[1:]
                self.field_names = self.field_names[1:]

    def clean_field_name(self, s):
        return re.sub('[^a-z0-9]+', '_', s.lower())

    def get_converter(self):
        return self.converter or RowConverter(
            self.database,
            has_header=self.has_header,
            sample_size=self.sample_size)

    def analyze_csv(self):
        converter = self.get_converter()
        header, rows = converter.extract_rows(
            self.file_or_name,
            **self.reader_kwargs)
        if rows:
            self.fields = converter.analyze(rows)
        else:
            self.fields = [converter.default.field() for _ in header]
        if not self.field_names:
            self.field_names = [self.clean_field_name(col) for col in  header]

    def get_model_class(self, field_names, fields):
        if self.model:
            return self.model
        attrs = dict(zip(field_names, fields))
        if 'id' not in attrs:
            attrs['_auto_pk'] = PrimaryKeyField()
        elif isinstance(attrs['id'], IntegerField):
            attrs['id'] = PrimaryKeyField()
        klass = type(self.db_table.title(), (Model,), attrs)
        klass._meta.database = self.database
        klass._meta.db_table = self.db_table
        return klass

    async def load(self):
        if not self.fields:
            self.analyze_csv()
        if not self.field_names and not self.has_header:
            self.field_names = [
                'field_%d' % i for i in range(len(self.fields))]

        #reader_obj = self.get_reader(self.file_or_name, **self.reader_kwargs)
        async with self.get_reader(self.file_or_name, **self.reader_kwargs) as reader:
            if not self.field_names:
                try:
                    row = await reader.__anext__()
                except AttributeError as te:
                    row = next(reader)
                except:
                    raise
                self.field_names = [self.clean_field_name(col) for col in row]
            elif self.has_header:
                try:
                    row = await reader.__anext__()
                except AttributeError as te:
                    row = next(reader)
                except:
                    raise


            ModelClass = self.get_model_class(self.field_names, self.fields)

            #with self.database.transaction():
            await ModelClass.create_table(True)
            inserts = []
            if isinstance(reader,Reader):
                async for row in reader:
                    if not row:
                        continue
                    insert = {}
                    for field_name, value in zip(self.field_names, row):
                        if value:
                            insert[field_name] = value
                    if insert:
                        inserts.append(insert)
            else:
                for row in reader:
                    if not row:
                        continue
                    insert = {}
                    for field_name, value in zip(self.field_names, row):
                        if value:
                            insert[field_name] = value
                    if insert:
                        inserts.append(insert)
            await ModelClass.insert_many(inserts).execute()

        return ModelClass

async def aioload_csv(db_or_model, file_or_name, fields=None, field_names=None,
             has_header=True, sample_size=10, converter=None,
             db_table=None, pk_in_csv=False, **reader_kwargs):
    loader = Loader(
        db_or_model=db_or_model,
        file_or_name=file_or_name,
        fields=fields,
        field_names=field_names,
        has_header=has_header,
        sample_size=sample_size,
        converter=converter,
        db_table=db_table,
        pk_in_csv=pk_in_csv,
        **reader_kwargs)
    return await loader.load()

aioload_csv.__doc__ = Loader.__doc__



from codecs import open
from setuptools import setup, find_packages
from os import path

REQUIREMETS_DEV_FILE = 'requirements_dev.txt'
REQUIREMETS_TEST_FILE = 'requirements_test.txt'
REQUIREMETS_FILE = 'requirements.txt'
PROJECTNAME = 'aioorm'
VERSION = '0.0.5'
DESCRIPTION = 'a simple orm suport asyncio,fork of aiopeewee'
URL = 'https://github.com/Python-Tools/aioorm'
AUTHOR = 'hsz'
AUTHOR_EMAIL = 'hsz1273327@gmail.com'
LICENSE = 'Mozilla'
CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Topic :: Documentation :: Sphinx',
]
KEYWORDS = ['orm', 'asyncio', 'peewee']
PACKAGES = find_packages(exclude=['contrib', 'docs', 'test'])
ZIP_SAFE = False

HERE = path.abspath(path.dirname(__file__))
with open(path.join(HERE, 'README.rst'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()
REQUIREMETS_DIR = path.join(HERE,"requirements")

with open(path.join(REQUIREMETS_DIR, REQUIREMETS_FILE), encoding='utf-8') as f:
    REQUIREMETS = f.readlines()

with open(path.join(REQUIREMETS_DIR, REQUIREMETS_DEV_FILE), encoding='utf-8') as f:
    REQUIREMETS_DEV = f.readlines()

with open(path.join(REQUIREMETS_DIR, REQUIREMETS_TEST_FILE), encoding='utf-8') as f:
    REQUIREMETS_TEST = f.readlines()
setup(
    name=PROJECTNAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license=LICENSE,
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    packages=PACKAGES,
    include_package_data=True,
    install_requires=REQUIREMETS,
    extras_require={
        'dev': REQUIREMETS_DEV,
        'test': REQUIREMETS_TEST
    },

    zip_safe=ZIP_SAFE,
    data_files=[('requirements', ['requirements/requirements.txt', 'requirements/requirements_dev.txt', 'requirements/requirements_test.txt'])]
)



from .model import AioModel
from .mysql import AioMySQLDatabase
from .fields import AioManyToManyField
from .shortcuts import model_to_dict
from .postgresql import AioPostgreSQLDatabase
from .utils import AioDbFactory



from peewee import Model, ModelAlias

from .query import *


class AioModelAlias(ModelAlias):

    def select(self, *selection):
        if not selection:
            selection = self.get_proxy_fields()
        query = AioSelectQuery(self, *selection)
        if self._meta.order_by:
            query = query.order_by(*self._meta.order_by)
        return query

    def __iter__(self):
        raise NotImplementedError()


class AioModel(Model):

    def __iter__(self):
        raise NotImplementedError()

    @classmethod
    def alias(cls):
        return AioModelAlias(cls)

    @classmethod
    def select(cls, *selection):
        query = AioSelectQuery(cls, *selection)
        if cls._meta.order_by:
            query = query.order_by(*cls._meta.order_by)
        return query

    @classmethod
    def update(cls, __data=None, **update):
        fdict = __data or {}
        fdict.update([(cls._meta.fields[f], update[f]) for f in update])
        return AioUpdateQuery(cls, fdict)

    @classmethod
    def insert(cls, __data=None, **insert):
        fdict = __data or {}
        fdict.update([(cls._meta.fields[f], insert[f]) for f in insert])
        return AioInsertQuery(cls, fdict)

    @classmethod
    def insert_many(cls, rows, validate_fields=True):
        return AioInsertQuery(cls, rows=rows, validate_fields=validate_fields)

    @classmethod
    def insert_from(cls, fields, query):
        return AioInsertQuery(cls, fields=fields, query=query)

    @classmethod
    def delete(cls):
        return AioDeleteQuery(cls)

    @classmethod
    def raw(cls, sql, *params):
        return AioRawQuery(cls, sql, *params)

    @classmethod
    async def create(cls, **query):
        inst = cls(**query)
        await inst.save(force_insert=True)
        inst._prepare_instance()
        return inst

    @classmethod
    async def get(cls, *query, **kwargs):
        sq = cls.select().naive()
        if query:
            sq = sq.where(*query)
        if kwargs:
            sq = sq.filter(**kwargs)
        return await sq.get()

    @classmethod
    async def get_or_create(cls, **kwargs):
        defaults = kwargs.pop('defaults', {})
        query = cls.select()
        for field, value in kwargs.items():
            if '__' in field:
                query = query.filter(**{field: value})
            else:
                query = query.where(getattr(cls, field) == value)

        try:
            return await query.get(), False
        except cls.DoesNotExist:
            try:
                params = dict((k, v) for k, v in kwargs.items()
                              if '__' not in k)
                params.update(defaults)
                async with cls._meta.database.atomic():
                    return await cls.create(**params), True
            except IntegrityError as exc:
                try:
                    return await query.get(), False
                except cls.DoesNotExist:
                    raise exc

    @classmethod
    async def table_exists(cls):
        kwargs = {}
        if cls._meta.schema:
            kwargs['schema'] = cls._meta.schema
        tables = await cls._meta.database.get_tables(**kwargs)
        return cls._meta.db_table in tables

    @classmethod
    async def create_table(cls, fail_silently=False):
        if fail_silently and await cls.table_exists():
            return

        db = cls._meta.database
        pk = cls._meta.primary_key
        if db.sequences and pk is not False and pk.sequence:
            if not db.sequence_exists(pk.sequence):
                db.create_sequence(pk.sequence)

        await db.create_table(cls)
        await cls._create_indexes()

    @classmethod
    async def _create_indexes(cls):
        for field_list, is_unique in cls._index_data():
            await cls._meta.database.create_index(cls, field_list, is_unique)

    @classmethod
    async def _drop_indexes(cls, safe=False):
        for field_list, is_unique in cls._index_data():
            await cls._meta.database.drop_index(cls, field_list, safe)

    @classmethod
    async def drop_table(cls, fail_silently=False, cascade=False):
        await cls._meta.database.drop_table(cls, fail_silently, cascade)

    @classmethod
    async def truncate_table(cls, restart_identity=False, cascade=False):
        await cls._meta.database.truncate_table(cls, restart_identity, cascade)

    @classmethod
    def noop(cls, *args, **kwargs):
        return NoopSelectQuery(cls, *args, **kwargs)

    async def save(self, force_insert=False, only=None):
        field_dict = dict(self._data)
        if self._meta.primary_key is not False:
            pk_field = self._meta.primary_key
            pk_value = self._get_pk_value()
        else:
            pk_field = pk_value = None
        if only:
            field_dict = self._prune_fields(field_dict, only)
        elif self._meta.only_save_dirty and not force_insert:
            field_dict = self._prune_fields(
                field_dict,
                self.dirty_fields)
            if not field_dict:
                self._dirty.clear()
                return False

        self._populate_unsaved_relations(field_dict)
        if pk_value is not None and not force_insert:
            if self._meta.composite_key:
                for pk_part_name in pk_field.field_names:
                    field_dict.pop(pk_part_name, None)
            else:
                field_dict.pop(pk_field.name, None)
            rows = await (self.update(**field_dict)
                              .where(self._pk_expr())
                              .execute())
        elif pk_field is None:
            await self.insert(**field_dict).execute()
            rows = 1
        else:
            pk_from_cursor = await self.insert(**field_dict).execute()
            if pk_from_cursor is not None:
                pk_value = pk_from_cursor
            self._set_pk_value(pk_value)
            rows = 1
        self._dirty.clear()
        return rows

    async def delete_instance(self, recursive=False, delete_nullable=False):
        if recursive:
            dependencies = self.dependencies(delete_nullable)
            for query, fk in reversed(list(dependencies)):
                model = fk.model_class
                if fk.null and not delete_nullable:
                    await (model.update(**{fk.name: None})
                                .where(query)
                                .execute())
                else:
                    await model.delete().where(query).execute()
        return await self.delete().where(self._pk_expr()).execute()
